import math
import time

import pygame
from OpenGL import GL
import ode
import random

import screen
import ship
import rope
import menu_screen
import sprite
import data

key_mapping = {
	pygame.K_LEFT: 0,
	pygame.K_DOWN: 1,
	pygame.K_RIGHT: 2,
	pygame.K_UP: 3,
	}
	
camera_extra_space = 4; #units (ship = unit)
camera_rotation_speed = 15 #degees per second

music_list = ["ingame1.mid","ingame2.mid","ingame3.mid","ingame4.mid","ingame5.mid","ingame6.mid","ingame7.mid"]

class Level (screen.Screen):
	def __init__(self, network_object):
		super(Level,self).__init__()
		random.seed(time.time())
		pygame.mixer.music.stop()
		pygame.mixer.music.load(data.filepath("music",random.choice(music_list)))
		pygame.mixer.music.play(-1)
		lvlname = network_object.handle_level_name()
		self.level_sprite = sprite.SpriteImage.getSpriteImage(data.filepath("backgrounds","nighttime.png"))
		self.rope_sprite = sprite.SpriteImage.getSpriteImage(data.filepath("rope.png"))
		self.quit_sprite = sprite.SpriteImage.getSpriteImage(data.filepath("quit.png"))
		self.next_level_sprite = sprite.SpriteImage.getTextSpriteImage(
			"Press SPACE to play another game..", font_size = 24)
		self.conn_lost_sprite = sprite.SpriteImage.getTextSpriteImage(
			"Connection lost!", font_size = 24)
		self.game_win_sprite = sprite.SpriteImage.getTextSpriteImage(
			"You won!", font_size = 52)
		self.game_lose_sprite = sprite.SpriteImage.getTextSpriteImage(
			"You lost!", font_size = 52)
		self.show_quit_sprite = False
		self.level_ended = False #only on server
		self.winners = None
		
		self.ships = []
		ship_type = network_object.get_ship_type(0)
		self.ships.append(ship.Ship(ship_type, [+4,+2.9], +math.pi/2,[+8,0]))
		ship_type = network_object.get_ship_type(1)
		self.ships.append(ship.Ship(ship_type, [+4,-2.9], -math.pi/2,[-8,0]))
		ship_type = network_object.get_ship_type(2)
		self.ships.append(ship.Ship(ship_type, [-4,+2.9], +math.pi/2,[+8,0]))
		ship_type = network_object.get_ship_type(3)
		self.ships.append(ship.Ship(ship_type, [-4,-2.9], -math.pi/2,[-8,0]))
		for i in range(len(self.ships)):
			self.ships[i].index = i
		
		self.ode_walls = []
		self.ode_space = ode.Space()
		self.__load_level(lvlname)
		self.network_object = network_object
		self.client_id = self.network_object.get_client_id()
	
	def _get_camera_bounding_box(self,rotation):
		minx = +10000.0
		maxx = -10000.0
		miny = +10000.0
		maxy = -10000.0
		rotation *= math.pi/180
		for ship in self.ships:
			rotated_position_x = math.cos(rotation)*ship.position[0] - math.sin(rotation)*ship.position[1]
			rotated_position_y = math.sin(rotation)*ship.position[0] + math.cos(rotation)*ship.position[1]
			minx = min(minx, rotated_position_x)
			maxx = max(maxx, rotated_position_x)
			miny = min(miny, rotated_position_y)
			maxy = max(maxy, rotated_position_y)
		minx -= camera_extra_space;
		miny -= camera_extra_space;
		maxx += camera_extra_space;
		maxy += camera_extra_space;
		return minx,miny,maxx,maxy;
	
	def _get_camera_zoom(self,rotation):
		minx,miny,maxx,maxy = self._get_camera_bounding_box(rotation)
		needex_x_zoom = 800/(maxx-minx)
		needex_y_zoom = 600/(maxy-miny)
		zoom = min(needex_x_zoom,needex_y_zoom,128) #zoom level (ship of size 1 will be maximum 128 pixels on screen)
		return zoom
	
	def _get_camera_center(self):
		centerx,centery = 0,0
		for ship in self.ships:
			centerx += ship.position[0] / len(self.ships)
			centery += ship.position[1] / len(self.ships)
		return centerx,centery
	
	def draw_background(self):
		GL.glPushMatrix()
		GL.glColor3f(1.0,1.0,1.0)
		GL.glScalef(25.0/self.level_sprite.size[0],25.0/self.level_sprite.size[1],1)
		GL.glTranslatef(-self.level_sprite.size[0] * 3,-self.level_sprite.size[1] * 3,0)
		for i in range(6):
			for j in range(6):
				self.level_sprite.draw()
				GL.glTranslatef(self.level_sprite.size[0],0,0)
			GL.glTranslatef(-self.level_sprite.size[0] * 6,self.level_sprite.size[1],0)
		GL.glPopMatrix()
	
	def draw(self):
		#drawing the level:
		GL.glPushMatrix()
		rotation = time.time()%360 * camera_rotation_speed
		centerx,centery = self._get_camera_center()
		zoom = self._get_camera_zoom(rotation)
		GL.glScalef(zoom,zoom,1)
		GL.glRotatef(rotation,0,0,1)
		GL.glTranslatef(-centerx,-centery,0)
		
		sprite.set_color((1,1,1,1))
		GL.glColor3f(1,1,1)
		self.draw_background()
		
		GL.glColor3f(0.5,0.5,0.5)
		for wall in self.ode_walls:
			GL.glRectf(*wall.coords)
		GL.glColor3f(1,1,1)
		
		rope.draw(self.ships[0], self.ships[1],self.rope_sprite)
		rope.draw(self.ships[2], self.ships[3],self.rope_sprite)
		for ship in self.ships:
			ship.draw()
		
		GL.glPopMatrix()
		
		#drawing the 2d overlay
		GL.glPushMatrix()
		GL.glTranslatef(-400,-300,0)
		
		sprite.set_color((1,1,1,1))
		GL.glColor3f(1,1,1)
		
		if self.winners != None and self.client_id in self.winners:
			GL.glPushMatrix()
			GL.glTranslatef(
				(800-self.game_win_sprite.size[0])/2,
				(600-self.game_win_sprite.size[1])/2,0)
			self.game_win_sprite.draw()
			GL.glPopMatrix()
		
		if self.winners != None and self.client_id not in self.winners:
			GL.glPushMatrix()
			GL.glTranslatef(
				(800-self.game_lose_sprite.size[0])/2,
				(600-self.game_lose_sprite.size[1])/2,0)
			self.game_lose_sprite.draw()
			GL.glPopMatrix()
		
		if self.show_quit_sprite:
			GL.glPushMatrix()
			GL.glTranslatef(
				(800-self.quit_sprite.size[0])/2,
				(600-self.quit_sprite.size[1])/2,0)
			self.quit_sprite.draw()
			GL.glPopMatrix()
			
		if self.level_ended:
			GL.glPushMatrix()
			GL.glTranslatef(
				(800-self.next_level_sprite.size[0])/2,
				480,0)
			self.next_level_sprite.draw()
			GL.glPopMatrix()
			
		if self.network_object == None:
			GL.glPushMatrix()
			GL.glTranslatef(
				(795-self.conn_lost_sprite.size[0]),
				5,0)
			self.conn_lost_sprite.draw()
			GL.glPopMatrix()
		
		GL.glPopMatrix()
	
	def ode_near(self, args, geom1, geom2):
		contacts = ode.collide(geom1,geom2)
		if (len(contacts)) == 0:
			return False
		col_ship,time_passed = args
		obj_geom = geom2
		px,py,pz = obj_geom.getPosition()
		lx,ly,lz = obj_geom.getLengths()
		col_ship.position = col_ship.last_position[:]
		if (abs(col_ship.last_position[0] - px) > lx/2.0):
			col_ship.velocity[0] *= -1
			col_ship.do_frame(time_passed)
		elif (abs(col_ship.last_position[1] - py) > ly/2.0):
			col_ship.velocity[1] *= -1
			col_ship.do_frame(time_passed)
		else:
			col_ship.velocity[0] *= -1
			col_ship.velocity[1] *= -1
			col_ship.do_frame(time_passed)

	def end_game(self, dead_ship):
		if self.network_object != None:
			self.network_object.handle_ship_dead(dead_ship,self)
	
	def do_frame(self, time_passed):
		if (self.network_object != None):
			try:
				self.network_object.do_frame(self, time_passed)
			except Exception, e:
				print "Error: %s\n" % e
				print "Removing network object\n"
				self.network_object = None
		for ship in self.ships:
			ship.do_frame(time_passed)
			ship_geom = ode.GeomBox(self.ode_space,(1.0/32.0,1.0/32,1))
			ship_geom.setPosition((ship.position[0],ship.position[1],0))
			self.ode_space.collide((ship,time_passed),self.ode_near)
			self.ode_space.remove(ship_geom)
		for ship in self.ships[:2]:
			if rope.cross(self.ships[2],self.ships[3],ship):
				self.end_game(ship)
		for ship in self.ships[2:]:
			if rope.cross(self.ships[0],self.ships[1],ship):
				self.end_game(ship)
		rope.pull(self.ships[0], self.ships[1], time_passed)
		rope.pull(self.ships[2], self.ships[3], time_passed)
	
	def handle_event(self, event):
		#KEYDOWN	unicode, key, mod
		#KEYUP		key, mod
		if event.type == pygame.KEYDOWN or event.type == pygame.KEYUP:
			if key_mapping.has_key(event.key):
				mapping = key_mapping[event.key]
				self.ships[self.client_id].keys_pressed[mapping] = (event.type == pygame.KEYDOWN)
				if self.network_object != None:
					self.network_object.send_updates(self.client_id, mapping, event.type == pygame.KEYDOWN)
		if event.type == pygame.KEYDOWN:
			if event.key == pygame.K_ESCAPE:
				self.show_quit_sprite = not self.show_quit_sprite
			if event.key == pygame.K_RETURN:
				if self.show_quit_sprite == True:
					self.next_level = menu_screen.MenuScreen()
			if event.key == pygame.K_SPACE:
				if self.level_ended: #on server only
					self.network_object.next_level(self)
	
	def __load_level(self, lvlname):
		fh = open(data.filepath("levels","%s.lvl" % lvlname),"rt")
		llist = fh.readlines()
		fh.close()
		for line in llist:
			if "" == line.strip():
				continue
			tup = line.strip().split(" ")
			tup = tuple(map(lambda x:float(x),tup))
			lens = (abs(tup[2]-tup[0]),abs(tup[3]-tup[1]),1)
			wall = ode.GeomBox(self.ode_space,lengths=lens)
			wall.coords = tup
			wall.setPosition(((tup[0] + tup[2])/2.0,(tup[1] + tup[3])/2.0,0))
			self.ode_walls.append(wall)
